  require 'logger'

class ExtendedLogger
  module Controls
    module ColorSchemes
      def self.example
        ColorScheme.build(
          :verbose => { :fg => :black, :bg => :white },
          :normal => { :fg => :white, :bg => :black },
        )
      end

      module None
        def self.example
          ColorScheme::None
        end
      end
    end

    module Formatters
      module Plain
        def self.example
          -> message do
            "#{message.level} -- #{message.prose}"
          end
        end
      end
    end

    module Label
      def self.example
        'SomeClass'
      end
    end

    module Levels
      def self.ruby_logger
        Set.ruby_logger.get :info
      end

      def self.example label=nil
        label ||= :normal
        Set.get label
      end

      module Names
        def self.ruby_logger
          %i(debug info warn error fatal)
        end

        def self.example
          %i(verbose normal)
        end
      end

      module Set
        def self.get name
          example[name]
        end

        def self.ruby_logger
          names = Names.ruby_logger
          Level::Set.build names
        end

        def self.example
          names = Names.example
          Level::Set.build names, :default => :normal
        end
      end
    end

    class Logger < ExtendedLogger.define Levels::Set.example, :format => Formatters::Plain.example
      def self.example
        device = StringIO.new

        instance = build device
        instance.clock = Time::Clock
        instance
      end

      class Color < ExtendedLogger.define Levels::Set.example, :color_scheme => ColorSchemes.example, :format => Formatters::Plain.example
        def self.example
          device = StringIO.new

          instance = build device
          instance.clock = Time::Clock
          instance
        end
      end

      class Ruby < ExtendedLogger.define
        def self.example
          device = StringIO.new

          instance = build device
          instance.clock = Time::Clock
          instance
        end
      end
    end

    module Messages
      def self.example prose: nil, level: nil, label: nil, time: nil
        prose ||= Prose.example
        level ||= Levels.example
        time ||= Time.reference

        if prose.is_a? Proc
          block = prose
        else
          block = ->{ prose }
        end

        Message.new block, level, time, label
      end

      module Label
        def self.example label=nil
          label ||= Control::Label.example

          Messages.example label: label
        end
      end

      module RubyLogger
        def self.example label=nil
          level = Level.new :info

          Messages.example level: level, label: label
        end
      end
    end

    module Output
      def self.example prose=nil
        prose ||= Prose.example
        "NORMAL -- #{prose}\n"
      end

      module Colorized
        def self.example prose=nil
          prose ||= Prose.example
          colorized_prose = "\e[1;37;40m#{prose}\e[0m"

          Output.example colorized_prose
        end
      end

      module MultipleLines
        def self.example
          "NORMAL -- Multiple\nNORMAL -- line\nNORMAL -- message\n"
        end
      end

      module RubyLogger
        def self.log_format
          @log_format ||= ::Logger::Formatter.new
        end

        def self.example progname=nil
          severity = 'INFO'
          time = Time::Clock.now
          message = Prose.example

          log_format.(severity, time, progname, message)
        end
      end

      module RuleApplied
        def self.example
          prose = Prose.example
          prose = prose.upcase

          Output.example prose
        end
      end
    end

    module Prose
      def self.example
        SingleLine.example
      end

      module EscapeCodes
        def self.example
          "\t\r"
        end
      end

      module MultipleLines
        def self.example
          "Multiple\nline\nmessage"
        end
      end

      module SingleLine
        def self.example
          "Some log message"
        end
      end
    end

    module Rule
      class Example
        include ExtendedLogger::Rule

        def call message
          message.prose = message.prose.upcase
          message
        end
      end
    end

    module Sink
      def self.example device=nil, format: nil, rules: nil
        format ||= Controls::Formatters::Plain.example
        device ||= StringIO.new
        levels = Controls::Levels::Set.example

        ExtendedLogger::Sink.build(
          device,
          format: format,
          levels: levels,
          rules: rules,
        )
      end
    end

    module Time
      module Clock
        def self.now
          Time.reference
        end
      end

      def self.reference
        ::Time.new 2000
      end
    end
  end
end
